package link;

import java.net.*;
import java.io.*;
import java.util.*;
import debug.*;
import timer.*;
import structure.*;

/***************************************************************** 
 *****************************************************************/
   /************************************************************
    ************************************************************/

/***************************************************************** 
   A link where the simulation model is managed by bouncing every
   packet off the echo server of a remote host using UDP.
 *****************************************************************/
public class EchoLink extends TimerClient
                      implements Link, Runnable
{
   /************************************************************
      Create a new link with the specified parameters

      @param name the debug name for this link
      
      @param mtu the maximum transmission unit, in bytes
      @param host the host name to bounce off
    ************************************************************/
   public EchoLink (String name, int mtu, String host, 
                    double bandwidthEst, double latencyEst)
   {
      this.name = name;
      this.mtu = mtu;

      this.bandwidthEst = bandwidthEst;
      this.latencyEst = latencyEst;

      nodes = new Vector ();

      try
         {
         if (host.equals ("localhost"))
            addr = InetAddress.getLocalHost ();
         else
            addr = InetAddress.getByName (host);
         }
      catch (Exception e)
         {
         System.out.println ("ERROR resolving hostname " + addr);
         System.exit (1);
         }

      try
         {
         socket = new DatagramSocket ();
         }
      catch (Exception e)
         {
         System.out.println ("ERROR opending UDP socket");
         System.exit (1);
         }

      Thread t = new Thread (this);
      t.start ();
   }

   /************************************************************
      Convert to a string for printing
    ************************************************************/
   public String toString ()
   {
      return "EchoLink: " + name;
   }

   private String name;
   private int mtu;

   private double bandwidthEst;
   private double latencyEst;

   private InetAddress addr;
   private DatagramSocket socket;

   private Vector nodes;    // all connected nodes

   /************************************************************
      Get the list of connected nodes

      @return an Enumeration of the currently connected nodes.
    ************************************************************/
   protected Enumeration ConnectedNodes ()
   {
      return nodes.elements ();
   }

   /************************************************************
      Transmit a packet on the link.  
      <p>
      Different nodes may be transmitting at the same time,
      but any particular node is limited to transmitting a
      single packet at a time.  Attempts for a single node
      to transmit multiple packets in parallel will block.

      @param p the packet to transmit
      @param n the node (router) sending the packet
      @see Link#TransmitPacket
    ************************************************************/
   public void TransmitPacket (LinkPacket p, Node n)
   {
      if (p.data.length > MTU ())
         return;

      EchoLinkArrivalRecord r = new EchoLinkArrivalRecord ();
      r.packet = p;

      // Do the transmission
      Node dest = OtherNode (n);
      for (int i=0; i<nodes.size (); ++i)
         if (dest == nodes.elementAt (i))
            {
            r.destinationOffset = i;
            break;
            }

      StructuredStorage s = new StructuredStorage ();
      s.ConvertFromObject (r);

      byte[] data = s.ConvertToByteArray ();

      DatagramPacket dp = new DatagramPacket (data, data.length, addr, 7);

      try
         {
         if (socket != null)
            socket.send (dp);
         else
            System.out.println ("null socket");
         }
      catch (IOException e)
         {
         System.out.println ("ERROR: IOException sending datagram.");
         System.exit (1);
         }      

      Debug.println ('e', "EchoLink: send packet to host " + addr);
   }


   /************************************************************
    ************************************************************/
   public void run ()
   {
      for (;;)
         {
         byte[] data = new byte[MTU () * 2];
         DatagramPacket dp = new DatagramPacket (data, data.length);

         try
            {
            socket.receive (dp);
            }
         catch (IOException e)
            {
            System.out.println ("ERROR: IOException receiving datagram.");
            System.exit (1);
            }

         Debug.println ('e', "EchoLink: packet arrived");

         StructuredStorage s = new StructuredStorage ();
         s.ConvertFromByteArray (dp.getData ());
         EchoLinkArrivalRecord r = (EchoLinkArrivalRecord) s.ConvertToObject ();

         Node node = (Node) nodes.elementAt (r.destinationOffset);

         synchronized (Timer.global)
            {
            node.PacketArrived (r.packet, this);
            }
         }
   }


   /************************************************************
      Get the maximum number of data bytes to transmit

      @return the maximum transmission unit, in bytes
      @see Link#MTU
    ************************************************************/
   public int MTU ()
   {
      return mtu;
   }

   /************************************************************
      Get the average bandwidth for this link

      @return the bandwidth, in Mb/s
      @see Link#Bandwidth
    ************************************************************/
   public double Bandwidth ()
   {
      return bandwidthEst;
   }

   /************************************************************
      Get the average latency for this link

      @return the latency, in ms
      @see Link#Latency
    ************************************************************/
   public double Latency ()
   {
      return latencyEst;
   }

   /************************************************************
      Get the names of all the other nodes connected to this 
      link (i.e. all the nodes who would receive a packet 
      that was transmitted on the link)

      @param n the node making the request (so that it can be
             excluded from the result)
      @return an array of the names of the other nodes
    ************************************************************/
   public String[] OtherNodeNames (Node n)
   {
      String[] retval = new String[1];
      retval[0] = "" + OtherNode (n);
      return retval;
   }

   /************************************************************
      Add a node, maintains the list of connected nodes.  At
      most two nodes may be connected at any time.
      @see Link#AddNode
    ************************************************************/
   public void AddNode (Node n)
   {
      if (nodes.size () < 2 && !nodes.contains (n))
         nodes.addElement (n);
   }

   /************************************************************
      Remove a node
      @see Link#RemoveNode
    ************************************************************/
   public void RemoveNode (Node n)
   {
      nodes.removeElement (n);
   }

   private Node OtherNode (Node n)
   {
      Node retval = null;

      if (nodes.size () == 2)
         {
         if (nodes.elementAt (0) == n)
            retval = (Node) nodes.elementAt (1);
         else
            retval = (Node) nodes.elementAt (0);
         }

      return retval;
   }
}

