package link;

import java.net.*;
import java.io.*;
import java.util.*;
import debug.*;
import timer.*;
import structure.*;

/***************************************************************** 
 *****************************************************************/
   /************************************************************
    ************************************************************/

/***************************************************************** 
   A simple point to point link that implements a randomized 
   delivery model.  No reordering is ever permitted, which does
   alter the simple latency model somewhat.
 *****************************************************************/
public class UDPLink extends TimerClient
                     implements Link, Runnable
{
   /************************************************************
      Create a new link with the specified parameters

      @param name the debug name for this link
      
      @param mtu the maximum transmission unit, in bytes

      @param dest the hostname of the destintation
      @param port the port to use (for send and receive)

      @param bandwidthEst an estimate of the link bandwidth, in Mb/s
      @param latencyEst an estimate of the link latency, in ms
    ************************************************************/
   public UDPLink (String name, int mtu, String dest, int port,
                   double bandwidthEst, double latencyEst)
   {
      this.name = name;
      this.mtu = mtu;

      this.bandwidthEst = bandwidthEst;
      this.latencyEst = latencyEst;

      node = null;

      remotePort = port;
      
      try
         {
         if (dest.equals ("localhost"))
            destAddr = InetAddress.getLocalHost ();
         else
            destAddr = InetAddress.getByName (dest);
         }
      catch (Exception e)
         {
         System.out.println ("ERROR resolving hostname " + dest);
         System.exit (1);
         }

      try
         {
         socket = new DatagramSocket ();
         }
      catch (Exception e)
         {
         System.out.println ("ERROR opending UDP socket on port " + port);
         System.exit (1);
         }

      localPort = socket.getLocalPort ();

      Thread t = new Thread (this);
      t.start ();
   }

   public UDPLink (String name, int mtu, int port,
                   double bandwidthEst, double latencyEst)
   {
      this.name = name;
      this.mtu = mtu;

      this.bandwidthEst = bandwidthEst;
      this.latencyEst = latencyEst;

      node = null;

      localPort = port;
      remotePort = -1;
      destAddr = null;

      try
         {
         socket = new DatagramSocket (localPort);
         }
      catch (Exception e)
         {
         System.out.println ("ERROR opending UDP socket on port " + port);
         System.exit (1);
         }

      Thread t = new Thread (this);
      t.start ();
   }

   /************************************************************
      Convert to a string for printing
    ************************************************************/
   public String toString ()
   {
      return "UDP: " + name;
   }

   private String name;
   private int mtu;

   private double bandwidthEst;
   private double latencyEst;

   private InetAddress destAddr;
   private int localPort;
   private int remotePort;
   private DatagramSocket socket;

   private Node node;

   /************************************************************
      Transmit a packet on the link.  Samples to find the transmit time,
      then waits for that time to simulate transmission.
      While one node is
      transmitting all others are blocked.  Fair queuing is not
      attempted.  The packets arrive at the other end after
      the latency (also sampled), with the constraint that they
      arrive in order and not exceed the arrival bandwidth.
      <p>
      Different nodes may be transmitting at the same time,
      but any particular node is limited to transmitting a
      single packet at a time.  Attempts for a single node
      to transmit multiple packets in parallel will block.

      @param p the packet to transmit
      @param n the node (router) sending the packet
      @see Link#TransmitPacket
    ************************************************************/
   public synchronized void TransmitPacket (LinkPacket p, Node n)
   {
      if (p.data.length > MTU ())
         return;

      StructuredStorage s = new StructuredStorage ();
      s.ConvertFromObject (p);

      byte[] data = s.ConvertToByteArray ();

      if (destAddr == null)
         return;

      DatagramPacket dp = new DatagramPacket (data, data.length,
                                              destAddr, remotePort);

      try
         {
         if (socket != null)
            socket.send (dp);
         }
      catch (IOException e)
         {
         System.out.println ("ERROR: IOException sending datagram.");
         System.exit (1);
         }

   }

   /************************************************************
      Get the maximum number of data bytes to transmit

      @return the maximum transmission unit, in bytes
      @see Link#MTU
    ************************************************************/
   public int MTU ()
   {
      return mtu;
   }

   /************************************************************
      Get the average bandwidth for this link

      @return the bandwidth, in Mb/s
      @see Link#Bandwidth
    ************************************************************/
   public double Bandwidth ()
   {
      return bandwidthEst;
   }

   /************************************************************
      Get the average latency for this link

      @return the latency, in ms
      @see Link#Latency
    ************************************************************/
   public double Latency ()
   {
      return latencyEst;
   }

   public String[] OtherNodeNames (Node n)
   {
      return new String[0];
   }

   /************************************************************
      Add a node, maintains the list of connected nodes.  At
      most two nodes may be connected at any time.
      @see Link#AddNode
    ************************************************************/
   public void AddNode (Node n)
   {
      if (node == null)
         node = n;
   }

   /************************************************************
      Remove a node
      @see Link#RemoveNode
    ************************************************************/
   public void RemoveNode (Node n)
   {
      node = null;
   }

   /************************************************************
    ************************************************************/
   public void run ()
   {
      for (;;)
         {
         byte[] data = new byte[MTU () * 2];
         DatagramPacket dp = new DatagramPacket (data, data.length);

         try
            {
            socket.receive (dp);
            }
         catch (IOException e)
            {
            System.out.println ("ERROR: IOException sending datagram.");
            System.exit (1);
            }

         if (destAddr == null)
            {
            destAddr = dp.getAddress ();
            remotePort = dp.getPort ();
            }

         StructuredStorage s = new StructuredStorage ();
         s.ConvertFromByteArray (dp.getData ());
         LinkPacket p = (LinkPacket) s.ConvertToObject ();

         synchronized (Timer.global)
            {
            if (node != null)
               node.PacketArrived (p, this);
            }
         }
   }
}
