package timer;

import java.io.*;
import java.util.*;
import debug.*;


/***************************************************************** 
   The real-time clock.  This is a large class, and
   the interface is divided into several sections:
   <ul>
   <li> <b>Clock interface</b>:    read the time
        <ul> 
        <li> <tt>public void CurrentSimTime ()</tt>
        </ul>

   <li> <b>Timeout interface</b>:  establish timeouts that will fire some
                                   time in the future
        <ul> 
        <li> <tt>public TimeoutID SetTimeout ()</tt>
        <li> <tt>public void CancelTimeout (TimeoutID)</tt>
        </ul>

   <li> <b>Wait interface</b>:     block the current thread until some
                                   time in the future.
        <ul>
        <li> <tt>public void WaitFor (TimerClient, long, TimeUnit)</tt>
        <li> <tt>public void WaitUntil (TimerClient, long, TimeUnit)</tt>
        </ul>
   </ul>
   
   The threading interaction between clients and the timer 
   is nontrivial.  The following rules of thumb should be 
   observed.
   <ul>
   <li> All client classes should <tt>extend</tt> the class
        <tt>TimerClient</tt>.  This identifies them as clients
        and gives them access to monitor routines that work
        correctly with the timer.

   <li> Clients which only want to set and receive timeouts should
        have little trouble.  They will need to override the 
        <tt>Timeout</tt> function with one of their own, and should
        generally avoid blocking or infinite processing inside
        this function.  If they do need to block, they should use
        either the modified routines in <tt>TimerClient</tt> or
        the <tt>WaitFor</tt> or <tt>WaitUntil</tt> functions in
        the timer.  They should avoid using the builtin Java
        blocking routines, as this typically leads to deadlock.

   <li> Clients which want to implement traditional multithreaded
        applications need to register and unregister all their
        threads with the timer.  Threads should remain registered
        throughout their entire lifetime, but <i>must</i> be 
        unregistered when they terminate.  Otherwise, these
        clients should observe the same rules as timeout clients.
   </ul>
 *****************************************************************/
public class Timer
{
   /************************************************************
    ************************************************************/

   /************************************************************
      The time unit seconds.  Time is maintained internally
      in microseconds, but these units allow more convenient
      specification of time periods in the public functions.
    ************************************************************/
   public static TimeUnit SEC = new TimeUnit (1000000);

   /************************************************************
      The time unit milliseconds
    ************************************************************/
   public static TimeUnit MS  = new TimeUnit (1000);

   /************************************************************
      The time unit microseconds
    ************************************************************/
   public static TimeUnit US  = new TimeUnit (1);


   /************************************************************
      The global timer.  Set up at program startup
    ************************************************************/
   public static Timer global = new Timer ();

   /************************************************************
      Sets stuff up and starts the event processing thread.  
      Clients are not allowed to create new timers.  They
      should use the global timer <tt>Timer.global</tt>

      @see Timer#global
    ************************************************************/
   Timer ()
   {
      pendingTimeouts = new Hashtable ();
      startTime = System.currentTimeMillis ();
   }

   /************************************************************
      Stops the timer, not really necessary.
    ************************************************************/
   public void Shutdown ()
   {
   }

   /************************************************************
      Return the current simulator time in microseconds
    ************************************************************/
   public long CurrentSimTime ()
   {
      return (System.currentTimeMillis () - startTime) * 1000;
   }




   /************************************************************
      SetTimeout:  set a timeout for the future

      @param client  the client to wake up when the time expires

      @param time    the number of time units from now that the
                     timeout will expire

      @param units   the units of time, one of
                     <ul>
                     <li>  <tt>Timer.SEC</tt>: seconds
                     <li>  <tt>Timer.MS</tt>: milliseconds
                     <li>  <tt>Timer.US</tt>: microseconds
                     </ul>

      @param info    the information about the timeout to pass
                     back to the client when the timeout expires

      @see TimerClient#Timeout
    ************************************************************/
   public synchronized TimeoutID SetTimeout (TimerClient client, 
                                             long time, 
                                             TimeUnit units,
                                             Object info)
   {
      long alarmTime = CurrentSimTime () + units.ConvertToUS (time);

      TimeoutID id = new TimeoutID ();
      TimeoutThread t = new TimeoutThread (id, client, info, alarmTime);
      t.start ();

      pendingTimeouts.put (id, t);

      return id;
   }

   /************************************************************
     Given an ID from a previous SetTimeout,
             remove it from the queue if it is still pending.	  
    ************************************************************/
   public synchronized void CancelTimeout (TimeoutID id)
   {
      if (id == null)
         throw new NullPointerException ("Tried to cancel null timeout");
      
      if (pendingTimeouts == null)
         {
         System.out.println ("BAD: pendingTimeouts went null during execution");
         System.exit (1);
         }

      TimeoutThread t = (TimeoutThread) pendingTimeouts.get (id);
      if (t != null)
         {
         t.stop ();
         pendingTimeouts.remove (t);
         }
   }



   /************************************************************
      Block the current thread on behalf of the client
      for a specified number
      of time units.  
      The current thread must
      also hold the client's monitor, which will be released for
      the duration of the wait.
	
      @param client the client to wait on behalf of             
      @param time   time units to sleep
      @param units  units for time
      @throws IllegalMonitorStateException if the current thread
              does not hold the client's monitor
    ************************************************************/
   public void WaitFor (TimerClient client, long time, TimeUnit units) 
                                 throws IllegalMonitorStateException
   {
      long wakeupTime = CurrentSimTime () + units.ConvertToUS (time);
      WaitUntil (client, wakeupTime, US);
   }

   /************************************************************
      Block the current thread until a specified simulates time.
      The current thread must be registered,
      and must be currently blocking time.
      The current thread must
      also hold the client's monitor, which will be released for
      the duration of the wait.
	
      @param client the client to wait on behalf of             
      @param time   time units to sleep
      @param units  units for time
      @throws IllegalMonitorStateException if the current thread
              does not hold the client's monitor
    ************************************************************/
   public void WaitUntil (TimerClient client, long time, TimeUnit units)
                                    throws IllegalMonitorStateException
   {
      Thread current = Thread.currentThread ();

      long wakeupTime = units.ConvertToUS (time);

      while (CurrentSimTime () < wakeupTime)
         {
         Debug.println ('t', "In WaitUntil on client " + client + 
                             ", about to call Wait");
         client.Wait ();
         }
   }

   /************************************************************
      Tell the simulator that this thread cares about simulated
       time.  This allows the thread to use the function
      <tt>BlockSimTimeFor</tt>, <tt>UnblockSimTimeFor</tt> and
      <tt>SimTimeIsBlockedFor</tt>, as well as the monitor
      functions in <tt>TimerClient</tt>.
      <p>
      Simulated time advances only when all registered threads
      are not blocking sim time.  When first registered,
      a thread <i>is</i> blocking sim time,

      @param t the thread to register
      @throws IllegalTimerStateException if this thread is
            already registered.
    ************************************************************/
   public synchronized void RegisterThread (Thread t)
                                       throws IllegalTimerStateException
   {
   }

   /************************************************************
     Tell the simulator that this thread no
                     longer cares about simulated time.

      @param t the thread to register
      @throws IllegalTimerStateException if this thread is not
         currently registered
    ************************************************************/
   public synchronized void UnregisterThread (Thread t) 
                                       throws IllegalTimerStateException
   {
   }

   /************************************************************
      Force simulated time to stall until this thread unblocks
      it or unregisters.  Threads should avoid sleeping while
      blocking sim time.  
      <p>
      In general it will probably not be
      necessary for any client to call this function, but it
      is available just in case.

      @param t the thread to block on behalf of
      @throws IllegalTimerStateException if this thread is not
         currently registered or is already blocking sim time
    ************************************************************/
   public synchronized void BlockSimTimeFor (Thread t)
                                        throws IllegalTimerStateException
   {
   }

   /************************************************************
      Remove the blocking constraint on this thread.
      <p>
      In general it will probably not be
      necessary for any client to call this function, but it
      is available just in case.

      @param t the thread to unblock on behalf of
      @throws IllegalTimerStateException if this thread is not
         currently registered or is not already blocking sim time
    ************************************************************/
   public synchronized void UnblockSimTimeFor (Thread t)
                                       throws IllegalTimerStateException
   {
   }


   /************************************************************
      Check whether this thread is blocking sim time.
      <p>
      In general it will probably not be
      necessary for any client to call this function, but it
      is available just in case.

      @param t the thread to check
      @throws IllegalTimerStateException if this thread is not
         currently registered.
      @return <tt>true</tt> if the thread is blocking sim time.
    ************************************************************/
   public synchronized boolean SimTimeIsBlockedFor (Thread t)
                                        throws IllegalTimerStateException
   {
      return false;
   }


   private long startTime;
   private Hashtable pendingTimeouts;
}




//-----------------------------------------------------------------
// TimeoutCallbackThread: a thread to run timeout processing 
//                        in the background.  This allows
//                        timeouts to run without holding the
//                        Timer monitor, which is desireable for
//                        deadlock avoidance.
//-----------------------------------------------------------------
class TimeoutThread extends Thread
{
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // Set up the thread
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   public TimeoutThread (TimeoutID id, TimerClient client, 
                         Object info, long wakeupTime)
   {
      this.id = id;
      this.client = client;
      this.info = info;

      sleepFor = wakeupTime - System.currentTimeMillis () * 1000;
      if (sleepFor < 0)
         sleepFor = 0;
   }

   private long sleepFor;
   private TimeoutID id;
   private TimerClient client;
   private Object info;

   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // Run the timeout.  Note that we run registered to prevent 
   // time from advancing.
   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   public void run ()
   {
      Thread.currentThread ().setName ("Timeout run thread for client " + id.client);

      try
         {
         Thread.sleep (sleepFor);
         synchronized (Timer.global)
            {
            client.Timeout (info);
            }
         }
      catch (InterruptedException e)
         {
         System.out.println ("Timeout interrupted for '" + 
                              client + "' info '" + info + "'");
         }
   }
}
